#!/usr/bin/env python
# -*- coding:utf-8 -*-
#
#   Author  :   cold
#   E-mail  :   wh_linux@126.com
#   Date    :   14/04/11 17:28:15
#   Desc    :   将返回信息实例化
#
""" 对应WebQQ的数据结构抽象
"""
import logging

logger = logging.getLogger("twqq")


class UniqueIds(object):
    """ 唯一ID
    为每个好友/群/讨论组/群成员都保持一个唯一标识
    """
    T_FRI = 1   # 好友
    T_TMP = 0   # 临时(群成员)
    T_GRP = 4   # 群
    T_DIS = 3   # 讨论组

    _map = {}     # 分配的id 到 uin和type的对应关系
    _r_map = {}   # uin 和 id 的对应关系
    _last_id = 0  # 最后一个id, 保持唯一

    @classmethod
    def alloc(cls, uin, _type):
        """ 分配一个, uin 是webqq的唯一标识, _type 对应 组/好友/讨论组/群成员
        并返回这个 id
        """
        assert _type in [cls.T_FRI, cls.T_TMP, cls.T_GRP, cls.T_DIS]
        if uin in cls._r_map:
            return cls._r_map[uin]
        else:
            _id = cls._last_id
            cls._last_id += 1
            cls._map[_id] = (uin, _type)
            cls._r_map[uin] = _id
            return _id

    @classmethod
    def get(cls, _id):
        """ 根据 _id 获取 uin 和对应的类型
        """
        return cls._map.get(_id, (None, None))

    @classmethod
    def get_type(cls, uin):
        """ 根据 uin 判断该uin 的类型
        """
        return cls._map.get(cls._r_map.get(uin), (None, None))[1]

    @classmethod
    def get_id(cls, uin):
        return cls._r_map.get(uin)


class ObjectsBase(object):

    def __init__(self, **kw):
        for key, val in kw.items():
            setattr(self, key, val)


class GroupMInfo(ObjectsBase):

    """ 对应群成员信息中的 minfo 字段中的元素

    :param nick: 成员qq昵称
    :param province: 成员省份
    :param gender: 成员性别
    :param uin: uin
    :param contry: 国家
    :param city: 城市
    :param client_type: 客户端类型
    :param stat: 目测固定都给10
    :param mflag: 1 则是管理员, 0 是普通成员
    :param card: 群名片
    """

    def __init__(self, nick, province, gender, uin, country, city,
                 stat=None, client_type=None, mflag=None, card=None):
        self.nick = nick
        self.province = province
        self.gender = gender   # 性别
        self.uin = uin
        self.country = country
        self.city = city

        self.stat = stat
        self.client_type = client_type
        self.card = card
        self._id = UniqueIds.alloc(uin, UniqueIds.T_TMP)

    def is_manager(self):
        return True if self.mflag == 1 else False


class VipInfo(ObjectsBase):

    """ 对应群信息 vipinfo 字段

    :param vip_level: 会员等级
    :param u: 用户 uin
    :param is_vip: 是否是管理员
    """

    def __init__(self, vip_level, u, is_vip):
        self.vip_level = vip_level
        self.u = u
        self.is_vip = True if is_vip == 1 else False


class Group(ObjectsBase):
    def __init__(self, flag, name, gid, code, memo=None, fingermemo=None,
                 createtime=None, level=None, owner=None, option=None,
                 members=None):
        self.flag = flag
        self.name = name
        self.gid = gid
        self.code = code
        self.group = None

        self.memo = memo
        self.fingermemo = fingermemo
        self.createtime = createtime
        self.level = level
        self.owner = owner
        self.option = option
        self._uin_map = {}   # 群成员 uin 映射
        self._uin_name_map = {}  # 群成员昵称到uin的映射
        self._id = UniqueIds.alloc(code, UniqueIds.T_GRP)

    def set_group_detail(self, data):
        """ 设置组详细信息, 包括群成员信息, 等.
        """
        for kw in data.get("minfo", []):
            tmp = GroupMInfo(**kw)
            self._uin_name_map[tmp.nick] = tmp.uin
            self._uin_map[tmp.uin] = tmp

        for item in data.get("cards", []):
            uin = item.get("muin")
            if uin in self._uin_map:
                self._uin_map[uin].card = item.get("card")
                self._uin_name_map[item.get("card")] = uin
            else:
                logger.warn(u"card info {0} not in map: {1!r}"
                            .format(uin, item))

        for item in data.get("stats", []):
            uin = item.get("uin")
            if uin in self._uin_map:
                self._uin_map[uin].stat = item.get("stat")
                self._uin_map[uin].client_type = item.get("client_type")
            else:
                logger.warn(u"stats info {0} not in map: {1!r}"
                            .format(uin, item))

        for item in data.get("vipinfo", []):
            u = item.get("u")
            if u in self._uin_map:
                self._uin_map[u].vipinfo = item
            else:
                logger.warn(u"vip info {0} not in map: {1!r}"
                            .format(uin, item))

        self.set_detail_info(**data.get("ginfo", {}))

    def __iter__(self):
        """ 迭代返回成员的 uin
        """
        if hasattr(dict, "iterkeys"):
            return self._uin_map.iterkeys()
        else:
            return self._uin_map.keys()

    def set_detail_info(self, face, memo, fingermemo, code, createtime,
                        flag, level, name, gid, owner, option, members, **kw):
        """ 组成员信息中对应 ginfo 中的元素

        :param face: 群头像
        :param memo:
        :param class: 群类型
        :param fingermemo:
        :param createtime: 创建时间戳
        :param flag:
        :param level:
        :param name: 群名称
        :param gid: 群hash码
        :param owner: 群主 uin
        :param option:
        :param members: 群成员
        """
        self.face = face
        self.memo = memo
        self._class = kw["class"]
        self.fingermemo = self.fingermemo
        self.createtime = createtime
        self.flag = flag
        self.level = level
        self.name = name
        self.gid = gid
        self.owner = owner
        self.option = option

        for item in members:
            uin = item.get("muin")
            tmp = self._uin_map.get(uin)
            if tmp:
                tmp.mflag = item.get("mflag")

    def __repr__(self):
        return u"<Group {0} have {1} members, Level {2}>"\
            .format(self.name, len(self._uin_map.keys()), self.level)

    def __unicode__(self):
        return self.__repr__()

    def get_nickname(self, uin):
        """ 获取群成员的昵称
        """
        r = self._uin_map.get(uin)
        if r:
            return r.nick

    def get_cardname(self, uin):
        """ 获取群成员群名片
        """
        r = self._uin_map.get(uin)
        if r:
            return r.card

    def get_show_name(self, uin):
        """ 获取显示名, 群名片优先, 无则取昵称
        """
        r = self._uin_map.get(uin)
        if r:
            return r.card if r.card else r.nick

    def get_member_info(self, uin):
        return self._uin_map.get(uin)

    def is_manager(self, uin):
        """ 判断用户是否是群管理员
        """
        r = self._uin_map.get(uin)
        if r:
            return r.is_manager()


class GroupList(ObjectsBase):
    """ 组列表抽象
    """

    def __init__(self, data):
        self._gcode_map = {}
        self._gid_gcode_map = {}
        self._gcode_name_map = {}
        for kw in data.get("gnamelist", []):
            group = Group(**kw)
            gcode = kw.get("code")
            self._gcode_map[gcode] = group
            self._gid_gcode_map[kw.get("gid")] = gcode
            self._gcode_name_map[kw.get("name")] = gcode
        self.gmasklist = data.get("gmasklist", [])
        self.gnamelist = [Group(**kw) for kw in data.get("gnamelist", [])]
        self.gmarklist = data.get("gmarklist", [])

    def __repr__(self):
        return str([x.name for x in self.gnamelist])

    def __unicode__(self):
        return self.__repr__().decode("utf-8")

    def __iter__(self):
        """ 迭代返回群对象
        """
        if hasattr(dict, "itervalues"):
            return self._gcode_map.itervalues()
        else:
            return self._gcode_map.values()

    @property
    def groups(self):
        return [x for x in self._gcode_map.values()]

    def get_gcodes(self):
        return [x.code for x in self.gnamelist]

    def find_group(self, gcode):
        return self._gcode_map.get(gcode)

    def set_group_info(self, gcode, data):
        """ 设置群信息

        :param gcode: 组代码
        :param data: 数据
        """
        item = self.find_group(gcode)
        if item:
            item.set_group_detail(data)

    def get_members(self, gcode):
        """  获取指定组的成员信息
        """
        item = self.find_group(gcode)
        if item:
            return item._uin_map.values()

    def get_member(self, gcode, uin):
        """ 获取指定群成员的信息
        """
        item = self.find_group(gcode)
        if item:
            return item.get_member(uin)

    def get_group_name(self, gcode):
        item = self.find_group(gcode)
        if item:
            return item.name

    def get_member_nick(self, gcode, uin):
        item = self.find_group(gcode)
        if item:
            return item.get_show_name(uin)

    def get_gid(self, gcode):
        item = self.find_group(gcode)
        if item:
            return item.gid

    def get_gcode(self, gid):
        return self._gid_gcode_map.get(gid)


class DiscuMemInfo(ObjectsBase):

    """ 讨论组成员信息

    :param uin: uin
    :param nick: 昵称
    """

    def __init__(self, uin, nick, status=None, client_type=None):
        self.uin = uin
        self.nick = nick
        self.status = status
        self.client_type = client_type
        self._id = UniqueIds.alloc(uin, UniqueIds.T_TMP)


class Discu(ObjectsBase):

    """ 讨论组信息抽象

    :param did: 讨论组id
    :param name: 讨论组名称
    """

    def __init__(self, did, name, discu_owner=None, discu_name=None,
                 info_seq=None, mem_list=None):
        self._uin_map = {}
        self.did = did
        self.name = name
        self.discu_name = discu_name,
        self.discu_owner = discu_owner
        self.info_seq = info_seq
        self.mem_list = mem_list
        self._id = UniqueIds.alloc(did, UniqueIds.T_DIS)

    def set_detail(self, info, mem_status, mem_info):
        self.discu_name = info.get("discu_name")
        self.discu_owner = info.get("discu_owner")
        self.info_seq = info.get("info_seq")
        for item in mem_info:
            self._uin_map[item["uin"]] = DiscuMemInfo(**item)

        for item in mem_status:
            tmp = self._uin_map.get(item["uin"])
            if tmp:
                tmp.status = item.get("status")
                tmp.client_type = item.get("client_type")

    def get_mname(self, uin):
        item = self._uin_map.get(uin)
        if item:
            return item.nick


class DiscuList(ObjectsBase):

    """ 讨论组列表
    """

    def __init__(self, data):
        self._did_map = {}
        self._did_name_map = {}
        for item in data.get("dnamelist"):
            self._did_name_map[item["name"]] = item["did"]
            self._did_map[item["did"]] = Discu(item["did"], item["name"])

    @property
    def dids(self):
        return self._did_map.keys()

    @property
    def discus(self):
        return [x for x in self._did_map.values()]

    def get_name(self, did):
        r = self._did_map.get(did)
        if r:
            return r.name

    def set_detail(self, did, data):
        self._did_map[did].set_detail(**data)

    def get_did(self, name):
        return self._did_name_map.get(name)

    def get_mname(self, did, uin):
        r = self._did_map.get(did)
        if r:
            return r.get_mname(uin)


class FriendInfo(ObjectsBase):

    """ 好友信息抽象

    :param uin: 唯一标识
    :param face: 头像
    :param nick: 昵称
    :param birthday: 生日
    :param occpation:
    :param phone: 手机号
    :param allow:
    :param colleage: 大学
    :param uin: uin
    :param constel:
    :param blood: 血型
    :param homepage: 主页
    :param stat: 状态
    :param vip_info: 是否vip
    :param country: 国家
    :param city: 城市
    :param personal:
    :param shengxiao: 生肖
    :param email: 邮件
    :param client_type: 客户端类型
    :param province: 省份
    :param gender: 性别
    :param mobile: 手机
    :param markname: 备注名
    :param categories: 分类id
    :param account: QQ号
    """
    def __init__(self, uin, face, flag, nick, birthday=None,
                 occpation=None, phone=None, allow=None, colleage=None,
                 constel=None, blood=None, homepage=None, status=None,
                 vip_info=None, country=None, city=None, personal=None,
                 shengxiao=None, email=None, client_type=None, province=None,
                 gender=None, mobile=None, markname=None, categories=None,
                 account=None):
        self.face = face
        self.flag = flag
        self.nick = nick
        self.uin = uin
        self.birthda = birthday
        self.occpation = occpation
        self.phone = phone
        self.allow = allow
        self.colleage = colleage
        self.constel = constel
        self.blood = blood
        self.homepage = homepage
        self.status = status
        self.vip_info = vip_info
        self.country = country
        self.city = city
        self.personal = personal
        self.shengxiao = shengxiao
        self.email = email
        self.client_type = client_type
        self.province = province
        self.gender = gender
        self.mobile = mobile
        self.markname = markname
        self.categories = categories
        self.account = account
        self._id = UniqueIds.alloc(uin, UniqueIds.T_FRI)

    def set_markname(self, markname):
        self.markname = markname

    def set_categories(self, categories):
        self.categories = categories

    def set_vipinfo(self, vipinfo):
        self.vip_info = vipinfo

    def set_detail(self, **kw):
        for k, v in kw.items():
            setattr(self, k, v)


class FriendCate(ObjectsBase):

    """ 好友分类

    :param index: 索引
    :param sort: 排序
    :param name: 名称
    """
    def __repr__(self):
        return u"<Category [{0}]>".format(self.name)


class Friends(ObjectsBase):

    def __init__(self, data):
        self._uin_map = {}
        self._name_map = {}
        self._mark_uin_map = {}

        for item in data.get("info", {}):
            info = FriendInfo(**item)
            self._uin_map[info.uin] = info
            self._name_map[info.nick] = info.uin

        for item in data.get("friends", []):
            uin = item.get("uin")
            self._uin_map[uin].set_categories(item.get("categories"))

        for item in data.get("marknames", []):
            uin = item.get("uin")
            self._uin_map[uin].set_markname(item.get("markname"))
            self._mark_uin_map[item.get("markname")] = uin

        for item in data.get("vipinfo", []):
            uin = item.get("u")
            self._uin_map[uin].set_vipinfo(item)
        self.categories = [FriendCate(**kw)
                           for kw in data.get("categories", [])]
        self.vipinfo = [VipInfo(**kw) for kw in data.get("vipinfo", [])]

    def get_uin(self, name):
        return self._name_map.get(name)

    @property
    def info(self):
        return [self._uin_map[uin] for uin in self._uin_map]

    def __repr__(self):
        return u"<{0} Friends>".format(len(self.info))

    def get_nick(self, uin):
        """ 获取好友信息昵称
        """
        item = self._uin_map.get(uin)
        if item:
            return item.nick

    def get_markname(self, uin):
        """ 获取好友备注信息
        """
        item = self._uin_map.get(uin)
        if item:
            return item.markname

    def get_show_name(self, uin):
        m = self.get_markname(uin)
        if not m:
            m = self.get_nick(uin)
        return m

    def get_uin_from_mark(self, mark):
        return self._mark_uin_map.get(mark)

    def set_status(self, uin, status, client_type):
        item = self._uin_map.get(uin)
        if item:
            item.status = status
            item.client_type = client_type

    def set_account(self, uin, account):
        item = self._uin_map.get(uin)
        if item:
            item.account = account

    def get_account(self, uin):
        item = self._uin_map.get(uin)
        if item:
            return item.account



#!/usr/bin/env python
# -*- coding:utf-8 -*-
#
#   Author  :   cold
#   E-mail  :   wh_linux@126.com
#   Date    :   13/11/11 18:21:10
#   Desc    :
#
import os
import json
import time
import random
import inspect
import logging

import const

logger = logging.getLogger("twqq")


class WebQQRequest(object):
    METHOD_POST = "post"
    METHOD_GET = "get"

    hub = None
    url = None
    params = {}
    headers = {}
    method = METHOD_GET         # 默认请求为GET
    kwargs = {}
    ready = True

    def __init__(self, *args, **kwargs):
        self.delay = kwargs.pop("delay", 0)
        self.init(*args, **kwargs)

    def handle_exc(self, type, value, trace):
        pass


class LoginSigRequest(WebQQRequest):
    url = "https://ui.ptlogin2.qq.com/cgi-bin/login"

    def init(self):
        self.hub.wait()
        logger.info("获取 login_sig...")
        self.params = [("daid", self.hub.daid), ("target", "self"),
                       ("style", 5), ("mibao_css", "m_webqq"),
                       ("appid", self.hub.aid), ("enable_qlogin", 0),
                       ("no_verifyimg", 1),
                       ("s_url", "http://web2.qq.com/loginproxy.html"),
                       ("f_url", "loginerroralert"),
                       ("strong_login", 1), ("login_state", 10),
                       ("t", "20130723001")]

    def callback(self, resp, data):
        if not data:
            logger.warn(u"没有获取到 Login Sig, 重新获取")
            return self.hub.load_next_request(self)

        sigs = self.hub.SIG_RE.findall(resp.body)
        if len(sigs) == 1:
            self.hub.login_sig = sigs[0]
            logger.info(u"获取Login Sig: {0}".format(self.hub.login_sig))
        else:
            logger.warn(u"没有获取到 Login Sig, 后续操作可能失败")

        self.hub.load_next_request(CheckRequest())


class CheckRequest(WebQQRequest):

    """ 检查是否需要验证码
    """
    url = "http://check.ptlogin2.qq.com/check"

    def init(self):
        self.params = {"uin": self.hub.qid, "appid": self.hub.aid,
                       "u1": const.CHECK_U1, "login_sig": self.hub.login_sig,
                       "js_ver": 10040, "js_type": 0, "r": random.random()}
        self.headers.update({"Referer": const.CHECK_REFERER})

    def callback(self, resp, data):
        r, vcode, uin = eval("self.hub." + data.strip().rstrip(";"))
        logger.debug("R:{0} vcode:{1}".format(r, vcode))
        self.hub.clean()
        if int(r) == 0:
            logger.info("验证码检查完毕, 不需要验证码")
            password = self.hub.handle_pwd(r, vcode, uin)
            self.hub.check_code = vcode
            self.hub.load_next_request(BeforeLoginRequest(password))
        else:
            logger.warn("验证码检查完毕, 需要验证码")
            self.hub.require_check = True
            self.hub.load_next_request(VerifyCodeRequest(r, vcode, uin))


class VerifyCodeRequest(WebQQRequest):
    url = "https://ssl.captcha.qq.com/getimage"

    def init(self, r, vcode, uin):
        self.r, self.vcode, self.uin = r, vcode, uin
        self.params = [("aid", self.hub.aid), ("r", random.random()),
                       ("uin", self.hub.qid)]

    def callback(self, resp, data):
        self.hub.require_check_time = time.time()
        with open(self.hub.checkimg_path, 'wb') as f:
            f.write(resp.body)
        self.hub.unwait()

        self.hub.client.handle_verify_code(self.hub.checkimg_path, self.r,
                                           self.uin)


class BeforeLoginRequest(WebQQRequest):

    """ 登录前的准备
    """
    url = "https://ssl.ptlogin2.qq.com/login"

    def init(self, password):
        self.hub.unwait()
        self.hub.lock()
        self.params = [("u", self.hub.qid), ("p", password),
                       ("verifycode", self.hub.check_code),
                       ("webqq_type", 10), ("remember_uin", 1),
                       ("login2qq", 1),
                       ("aid", self.hub.aid), ("u1", const.BLOGIN_U1),
                       ("h", 1), ("action", '4-5-8246'), ("ptredirect", 0),
                       ("ptlang", 2052), ("from_ui", 1),
                       ("daid", self.hub.daid),
                       ("pttype", 1), ("dumy", ""), ("fp", "loginerroralert"),
                       ("mibao_css", "m_webqq"), ("t", 1), ("g", 1),
                       ("js_type", 0),
                       ("js_ver", 10040), ("login_sig", self.hub.login_sig)]
        referer = const.BLOGIN_R_REFERER if self.hub.require_check else\
            const.BLOGIN_REFERER
        self.headers.update({"Referer": referer})

    def ptuiCB(self, scode, r, url, status, msg, nickname=None):
        """ 模拟JS登录之前的回调, 保存昵称 """
        return scode, r, url, status, msg, nickname

    def get_back_args(self, data):
        blogin_data = data.decode("utf-8").strip().rstrip(";")
        logger.info(u"登录返回数据: {0}".format(blogin_data))
        return eval("self." + blogin_data)

    def check(self, scode, r, url, status, msg, nickname=None):
        self.hub.unlock()
        if int(scode) == 0:
            logger.info("从Cookie中获取ptwebqq的值")
            old_value = self.hub.ptwebqq
            try:
                val = self.hub.http.cookie['.qq.com']['/']['ptwebqq'].value
                self.hub.ptwebqq = val
            except:
                logger.error("从Cookie中获取ptwebqq的值失败, 使用旧值尝试")
                self.hub.ptwebqq = old_value
        elif int(scode) == 4:
            logger.error(msg)
            return False, self.hub.load_next_request(CheckRequest())
        else:
            logger.error(u"server response: {0}".format(msg.decode('utf-8')))
            return False, self.hub.load_next_request(CheckRequest())

        if nickname:
            self.hub.nickname = nickname.decode('utf-8')

        return True, url

    def callback(self, resp, data):
        if not data:
            logger.error("没有数据返回, 登录失败, 尝试重新登录")
            return self.hub.load_next_request(CheckRequest())

        args = self.get_back_args(resp.body)
        r, url = self.check(*args)
        if r:
            logger.info("检查完毕")
            self.hub.load_next_request(LoginRequest(url))


class LoginRequest(WebQQRequest):

    """ 登录前的准备
    """

    def init(self, url):
        logger.info("开始登录前准备...")
        self.url = url
        self.headers.update(Referer=const.LOGIN_REFERER)

    def callback(self, resp, data):
        if os.path.exists(self.hub.checkimg_path):
            os.remove(self.hub.checkimg_path)

        self.hub.load_next_request(Login2Request())


class Login2Request(WebQQRequest):

    """ 真正的登录
    """
    url = "http://d.web2.qq.com/channel/login2"
    method = WebQQRequest.METHOD_POST

    def init(self, relogin=False):
        self.relogin = relogin
        logger.info("准备完毕, 开始登录")
        self.headers.update(Referer=const.S_REFERER, Origin=const.D_ORIGIN)
        self.params = [("r", json.dumps({"status": "online",
                                         "ptwebqq": self.hub.ptwebqq,
                                         "passwd_sig": "",
                                         "clientid": self.hub.clientid,
                                         "psessionid": None})),
                       ("clientid", self.hub.clientid),
                       ("psessionid", "null")]

    def callback(self, resp, data):
        self.hub.require_check_time = None
        if not resp.body or not isinstance(data, dict):
            logger.error(u"没有获取到数据或数据格式错误, 登录失败:{0}"
                         .format(resp.body))
            self.hub.load_next_request(FirstRequest())
            return

        if data.get("retcode") != 0:
            logger.error("登录失败 {0!r}".format(data))
            self.hub.load_next_request(FirstRequest())
            return
        self.hub.vfwebqq = data.get("result", {}).get("vfwebqq")
        self.hub.psessionid = data.get("result", {}).get("psessionid")

        if not self.relogin:
            logger.info("登录成功, 开始加载好友列表")
            self.hub.load_next_request(FriendListRequest())
        else:
            logger.info("重新登录成功, 开始拉取消息")
            self.hub.start_poll()


class FriendListRequest(WebQQRequest):

    """ 加载好友信息
    """
    url = "http://s.web2.qq.com/api/get_user_friends2"
    method = WebQQRequest.METHOD_POST

    def init(self, first=True, manual=False):
        self.is_first = first
        self.manual = manual
        self.params = [("r", json.dumps({"h": "hello",
                                         "hash": self.hub._hash(),
                                         "vfwebqq": self.hub.vfwebqq}))]
        self.headers.update(Referer=const.S_REFERER)

    def callback(self, resp, data):
        if not resp.body and self.is_first:
            logger.error("加载好友信息失败, 重新开始登录")
            return self.hub.load_next_request(FirstRequest())
        if data.get("retcode") != 0 and self.is_first:
            logger.error("加载好友信息失败, 重新开始登录")
            return self.hub.load_next_request(FirstRequest())

        if isinstance(data, dict) and data.get("retcode") == 0:
            info = data.get("result", {})
            self.hub.set_friends(info)
            friends = self.hub.get_friends()
            logger.info(u"加载好友信息 {0!r}".format(friends))
            logger.debug(data)
            self.hub.load_next_request(FriendStatusRequest())
        else:
            logger.warn(u"加载好友列表失败: {0!r}".format(data))

        if not self.manual:
            self.hub.load_next_request(GroupListRequest())
            self.hub.load_next_request(DiscuListRequest())
            self.hub.load_next_request(FriendListRequest(delay=3600,
                                                         first=False))


FriendInfoRequest = FriendListRequest
import warnings
warnings.warn("In next version we will rename twqq.requests.FreindInfoRequest "
              "to twqq.requests.FriendListRequest")


class FriendStatusRequest(WebQQRequest):

    """ 获取在线好友状态
    """

    url = "https://d.web2.qq.com/channel/get_online_buddies2"

    def init(self):
        self.params = {"clientid": self.hub.clientid,
                       "psessionid": self.hub.psessionid,
                       "t": int(time.time() * 1000)}
        self.headers.update(Referer=const.D_REFERER)

    def callback(self, response, data):
        logger.info(u"加载好友状态信息: {0!r}".format(data))
        if isinstance(data, dict) and data.get("retcode") == 0:
            for item in data.get('result', []):
                self.hub.get_friends().set_status(**item)
        else:
            logger.warn(u"加载好友状态信息失败: {0!r}".format(data))


class GroupListRequest(WebQQRequest):

    """ 获取群列表
    """
    url = "http://s.web2.qq.com/api/get_group_name_list_mask2"
    method = WebQQRequest.METHOD_POST

    def init(self):
        self.params = {"r": json.dumps({"vfwebqq": self.hub.vfwebqq})}
        self.headers.update(Origin=const.S_ORIGIN)
        self.headers.update(Referer=const.S_REFERER)
        logger.info("获取群列表")

    def callback(self, resp, data):
        logger.debug(u"群信息 {0!r}".format(data))
        self.hub.set_groups(data.get("result", {}))
        groups = self.hub.get_groups()
        logger.info(u"群列表: {0!r}".format(groups))
        if not groups.gnamelist:
            self.hub.start_poll()

        for i, gcode in enumerate(groups.get_gcodes()):
            self.hub.load_next_request(GroupMembersRequest(gcode, i == 0))


class GroupMembersRequest(WebQQRequest):

    """ 获取群成员

    :param gcode: 群代码
    :param poll: 是否开始拉取信息和心跳
    :type poll: boolean
    """
    url = "http://s.web2.qq.com/api/get_group_info_ext2"

    def init(self, gcode, poll=False):
        self._poll = poll
        self._gcode = gcode
        self.params = [("gcode", gcode), ("vfwebqq", self.hub.vfwebqq),
                       ("cb", "undefined"), ("t", int(time.time()))]
        self.headers.update(Referer=const.S_REFERER)

    def callback(self, resp, data):
        if isinstance(data, dict) and data.get("retcode") == 0:
            logger.debug(u"获取群成员信息 {0!r}".format(data))
            members = data.get("result", {})
            groups = self.hub.get_groups()
            group = groups.find_group(self._gcode)
            group.set_group_detail(members)
            logger.debug(u"群详细信息: {0}".format(group))
        else:
            logger.warn(u"获取群成员信息失败 {0!r}".format(data))

        if self._poll:
            self.hub.start_poll()


class HeartbeatRequest(WebQQRequest):

    """ 心跳请求
    """
    url = "http://web.qq.com/web2/get_msg_tip"
    kwargs = dict(request_timeout=0.5, connect_timeout=0.5)

    def init(self):
        self.params = dict([("uin", ""), ("tp", 1), ("id", 0), ("retype", 1),
                            ("rc", self.hub.rc), ("lv", 3),
                            ("t", int(time.time() * 1000))])
        self.hub.rc += 1

    def callback(self, resp, data):
        logger.info("心跳...")


class PollMessageRequest(WebQQRequest):

    """ 拉取消息请求
    """
    url = "http://d.web2.qq.com/channel/poll2"
    method = WebQQRequest.METHOD_POST
    kwargs = {"request_timeout": 60.0, "connect_timeout": 60.0}

    def init(self):
        rdic = {"clientid": self.hub.clientid,
                "psessionid": self.hub.psessionid, "key": 0, "ids": []}
        self.params = [("r", json.dumps(rdic)),
                       ("clientid", self.hub.clientid),
                       ("psessionid", self.hub.psessionid)]
        self.headers.update(Referer=const.D_REFERER)
        self.headers.update(Origin=const.D_ORIGIN)
        self.ready = not self.hub.stop_poll

    def callback(self, resp, data):
        polled = False
        try:
            if not data:
                return

            if data.get("retcode") in [121, 120]:
                logger.error("获取登出消息, 尝试重新登录")
                self.hub.relogin()
                return

            logger.info(u"获取消息: {0!r}".format(data))
            self.hub.load_next_request(PollMessageRequest())
            polled = True
            self.hub.dispatch(data)
        except Exception as e:
            logger.error(u"消息获取异常: {0}".format(e), exc_info=True)
        finally:
            if not polled:
                self.hub.load_next_request(PollMessageRequest())


class SessGroupSigRequest(WebQQRequest):

    """ 获取临时消息群签名请求

    :param qid: 临时签名对应的qid(对应群的gid)
    :param to_uin: 临时消息接收人uin
    :param sess_reqeust: 发起临时消息的请求
    """

    url = "https://d.web2.qq.com/channel/get_c2cmsg_sig2"

    def init(self, qid, to_uin, sess_reqeust):
        self.sess_request = sess_reqeust
        self.to_uin = to_uin
        self.params = (("id", qid), ("to_uin", to_uin),
                       ("service_type", 0), ("clientid", self.hub.clientid),
                       ("psessionid", self.hub.psessionid), ("t", time.time()))
        self.headers.update(Referer=const.S_REFERER)

    def callback(self, resp, data):
        result = data.get("result")
        group_sig = result.get("value")
        if data.get("retcode") != 0:
            logger.warn(u"加载临时消息签名失败: {0}".format(group_sig))
            return

        logger.info(u"加载临时消息签名 {0} for {1}".format(group_sig, self.to_uin))
        self.hub.group_sig[self.to_uin] = group_sig
        self.sess_request.ready = True
        self.sess_request.init_params(group_sig)
        self.hub.load_next_request(self.sess_request)


class SessMsgRequest(WebQQRequest):

    """ 发送临时消息请求

    :param qid: 临时消息qid
    :param to_uin: 接收人 uin
    :param content: 发送内容
    """
    url = "https://d.web2.qq.com/channel/send_sess_msg2"
    method = WebQQRequest.METHOD_POST

    def init(self, qid, to_uin, content, style):
        self.to = to_uin
        self._content = content
        self.content = self.hub.make_msg_content(content, style)
        group_sig = self.hub.group_sig.get(to_uin)
        if not group_sig:
            self.ready = False
            self.hub.load_next_request(SessGroupSigRequest(qid, to_uin, self))
        else:
            self.init_params(group_sig)

    def init_params(self, group_sig):
        self.delay, self.number = self.hub.get_delay(self._content)
        self.params = (("r", json.dumps({"to": self.to, "group_sig": group_sig,
                                         "face": 549, "content": self.content,
                                         "msg_id": self.hub.msg_id,
                                         "service_type": 0,
                                         "clientid": self.hub.clientid,
                                         "psessionid": self.hub.psessionid})),
                       ("clientid", self.hub.clientid),
                       ("psessionid", self.hub.psessionid))

    def callback(self, resp, data):
        logger.info(u"发送给 {0} 临时消息成功: {1}".format(self.to, data))
        self.hub.consume_delay(self.number)


class GroupMsgRequest(WebQQRequest):

    """ 发送群消息

    :param group_uin: 群uin
    :param content: 消息内容
    """
    url = "http://d.web2.qq.com/channel/send_qun_msg2"
    method = WebQQRequest.METHOD_POST

    def init(self, group_uin, content, style):
        self.delay, self.number = self.hub.get_delay(content)
        self.gid = self.hub.get_group_id(group_uin)
        self.group_uin = group_uin
        self.source = content
        content = self.hub.make_msg_content(content, style)
        r = {"group_uin": self.gid, "content": content,
             "msg_id": self.hub.msg_id, "clientid": self.hub.clientid,
             "psessionid": self.hub.psessionid}
        self.params = [("r", json.dumps(r)),
                       ("psessionid", self.hub.psessionid),
                       ("clientid", self.hub.clientid)]
        self.headers.update(Origin=const.D_ORIGIN, Referer=const.D_REFERER)

    def callback(self, resp, data):
        logger.info(u"发送群消息 {0} 到 {1} 成功: {2}"
                    .format(self.source, self.group_uin, data))
        self.hub.consume_delay(self.number)


class DiscuListRequest(WebQQRequest):
    """ 获取讨论组列表
    """
    url = "http://s.web2.qq.com/api/get_discus_list"

    def init(self):
        self.params = {"clientid": self.hub.clientid,
                       "psessionid": self.hub.psessionid,
                       "vfwebqq": self.hub.vfwebqq,
                       "t": time.time() * 1000}
        self.headers.update(Referer=const.S_REFERER)

    def callback(self, resp, data):
        logger.info(u"获取讨论组列表: {0!r}".format(data))
        if data.get("retcode") == 0:
            self.hub.set_discu(data.get("result", {}))
            dids = self.hub.get_discu().dids
            for did in dids:
                self.hub.load_next_request(DiscuInfoRequest(did))


class QQNumberRequest(WebQQRequest):
    """ 查看好友QQ号码
    """
    url = "http://s.web2.qq.com/api/get_friend_uin2"

    def init(self, uin):
        self.params = {"code": "", "t": time.time() * 1000,
                       "tuin": uin, "type": 0,
                       "verifysession": "",
                       "vfwebqq": self.hub.vfwebqq}
        self.headers.update(Referer=const.S_REFERER)

    def callback(self, response, data):
        if isinstance(data, dict) and data.get("retcode") == 0:
            r = data.get('result')
            uin = r.get("uin")
            account = r.get("account")
            self.hub.get_friends().set_account(uin, account)
        else:
            logger.warn(u"获取好友QQ号码失败: {0!r}".format(data))


class DiscuInfoRequest(WebQQRequest):
    """ 获取讨论组详细信息

    :param did: 讨论组id
    """
    url = "https://d.web2.qq.com/channel/get_discu_info"

    def init(self, did):
        self.params = {"clientid": self.hub.clientid,
                       "did": did,
                       "psessionid": self.hub.psessionid,
                       "vfwebqq": self.hub.vfwebqq,
                       "t": time.time() * 1000}
        self.headers.update(Referer=const.S_REFERER)
        self._did = did

    def callback(self, resp, data):
        if data.get("retcode") == 0:
            discu = self.hub.get_discu()
            logger.info(u"获取讨论组 {0} 的详细信息: {1}"
                        .format(discu.get_name(self._did), data))
            discu.set_detail(self._did, data.get("result", {}))


class DiscuMsgRequest(WebQQRequest):
    """ 发送讨论组消息请求
    :param did: 讨论组id
    :param content: 消息内容
    """

    url = "https://d.web2.qq.com/channel/send_discu_msg2"
    method = WebQQRequest.METHOD_POST

    def init(self, did, content, style):
        self.delay, self.number = self.hub.get_delay(content)
        self.did = did
        self.source = content
        content = self.hub.make_msg_content(content, style)
        r = {"did": did, "content": content,
             "msg_id": self.hub.msg_id, "clientid": self.hub.clientid,
             "psessionid": self.hub.psessionid}
        self.params = [("r", json.dumps(r)),
                       ("psessionid", self.hub.psessionid),
                       ("clientid", self.hub.clientid)]
        self.headers.update(Referer=const.D_REFERER)

    def callback(self, resp, data):
        logger.info(u"发送讨论组消息 {0} 到 {1} 成功: {2}"
                    .format(self.source, self.did, data))
        self.hub.consume_delay(self.number)


class BuddyMsgRequest(WebQQRequest):

    """ 好友消息请求

    :param to_uin: 消息接收人
    :param content: 消息内容
    :param callback: 消息发送成功的回调
    """
    url = "http://d.web2.qq.com/channel/send_buddy_msg2"
    method = WebQQRequest.METHOD_POST

    def init(self, to_uin, content, style):
        self.to_uin = to_uin
        self.source = content
        self.content = self.hub.make_msg_content(content, style)
        r = {"to": to_uin, "face": 564, "content": self.content,
             "clientid": self.hub.clientid, "msg_id": self.hub.msg_id,
             "psessionid": self.hub.psessionid}
        self.params = [("r", json.dumps(r)), ("clientid", self.hub.clientid),
                       ("psessionid", self.hub.psessionid)]
        self.headers.update(Origin=const.D_ORIGIN)
        self.headers.update(Referer=const.S_REFERER)

        self.delay, self.number = self.hub.get_delay(content)

    def callback(self, resp, data):
        logger.info(u"发送好友消息 {0} 给 {1} 成功: {2}"
                    .format(self.source, self.to_uin, data))

        self.hub.consume_delay(self.number)


class SetSignatureRequest(WebQQRequest):

    """ 设置个性签名请求

    :param signature: 签名内容
    """
    url = "http://s.web2.qq.com/api/set_long_nick2"
    method = WebQQRequest.METHOD_POST

    def init(self, signature):
        self.params = (("r", json.dumps({"nlk": signature,
                                         "vfwebqq": self.hub.vfwebqq})),)
        self.headers.update(Origin=const.S_ORIGIN)
        self.headers.update(Referer=const.S_REFERER)

    def callback(self, resp, data):
        logger.info(u"设置签名成功: {0}".format(data))


class AcceptVerifyRequest(WebQQRequest):

    """ 同意好友添加请求

    :param uin: 请求人uin
    :param qq_num: 请求人QQ号
    """
    url = "http://s.web2.qq.com/api/allow_and_add2"
    method = WebQQRequest.METHOD_POST

    def init(self, uin, qq_num, markname=""):
        self.uin = uin
        self.qq_num = qq_num
        self.markname = markname
        self.params = [("r", "{\"account\":%d, \"gid\":0, \"mname\":\"%s\","
                        " \"vfwebqq\":\"%s\"}" % (qq_num, markname,
                                                  self.hub.vfwebqq)), ]
        self.headers.update(Origin=const.S_ORIGIN)
        self.headers.update(Referer=const.S_REFERER)

    def callback(self, resp, data):
        if data.get("retcode") == 0:
            logger.info(u"添加 {0} 成功".format(self.qq_num))
            # if self.markname:
            #     self.hub.mark_to_uin[self.markname] = self.uin
        else:
            logger.info(u"添加 {0} 失败".format(self.qq_num))


class FileRequest(WebQQRequest):
    """ 下载传送的文件

    :param guid: 文件名
    :param lcid: 会话id
    :param to_uin: 发送人uin
    """
    url = "http://d.web2.qq.com/channel/get_file2"

    def init(self, guid, lcid, to, callback=None):
        self.params = {"clientid": self.hub.clientid,
                       "psessionid": self.hub.psessionid,
                       "count": 1, "time": int(time.time() * 1000),
                       "guid": guid, "lcid": lcid, "to": to}
        self.headers.update(Referer="http://web2.qq.com/webqq.html")
        self.headers.pop("Origin", None)
        self.fname = guid
        self._cb = callback

    def callback(self, response, data):
        """ 应该在客户端通过::

            @register_request_handler(FileRequest)
            def callback(resp, data):
                pass

        来实现文件的保存
        """
        if self._cb:
            self._cb(self.fname, response.body)


class LogoutRequset(WebQQRequest):
    """ 登出请求
    """
    url = "https://d.web2.qq.com/channel/logout2"

    def init(self):
        self.params = {"clientid": self.hub.clientid, "ids": "",
                       "psessionid": self.hub.psessionid,
                       "t": int(time.time() * 1000)}
        self.headers.update(Referer=const.D_REFERER)

    def callback(self, resp, data):
        if data.get("retcode") == 0:
            logger.info(u"登出成功")


FirstRequest = LoginSigRequest


def _register_message_handler(func, args_func, msg_type="message"):
    """ 注册成功消息器

    :param func: 处理器
    :param args_func: 产生参数的处理器
    :param mst_type: 处理消息的类型
    """
    func._twqq_msg_type = msg_type
    func._args_func = args_func
    return func


def group_message_handler(func):
    """ 装饰处理群消息的函数

    处理函数应接收5个参数:

        nickname        发送消息的群昵称
        content         消息内容
        group_code      群代码
        from_uin        发送人的uin
        source          消息原包
    """

    def args_func(self, message):
        value = message.get("value", {})
        gcode = value.get("group_code")
        uin = value.get("send_uin")
        contents = value.get("content", [])
        content = self.handle_qq_msg_contents(uin, contents, gcode)
        uname = self.get_group_member_nick(gcode, uin)
        return uname, content, gcode, uin, message

    return _register_message_handler(func, args_func, "group_message")


def buddy_message_handler(func):
    """ 装饰处理好友消息的函数

    处理函数应接收3个参数:

        from_uin         发送人uin
        content          消息内容
        source           消息原包
    """

    def args_func(self, message):
        value = message.get("value", {})
        from_uin = value.get("from_uin")
        contents = value.get("content", [])
        content = self.handle_qq_msg_contents(from_uin, contents)
        return from_uin, content, message
    return _register_message_handler(func, args_func, "message")


def kick_message_handler(func):
    """ 装饰处理下线消息的函数

    处理函数应接收1个参数:

        source      消息原包
    """

    def args_func(self, message):
        return message,
    return _register_message_handler(func, args_func, "kick_message")


def sess_message_handler(func):
    """ 装饰处理临时消息的函数

    处理函数应接收3个参数:

        id              获取组签名的id
        from_uin        发送人uin
        content         消息内容
        source          消息原包
    """

    def args_func(self, message):
        value = message.get("value", {})
        id_ = value.get("id")
        from_uin = value.get("from_uin")
        contents = value.get("content", [])
        content = self.handle_qq_msg_contents(from_uin, contents)
        return id_, from_uin, content, message

    return _register_message_handler(func, args_func, "sess_message")


def system_message_handler(func):
    """ 装饰处理系统消息的函数

    处理函数应接手4个参数:

        type        消息类型
        from_uin    产生消息的人的uin
        account     产生消息的人的qq号
        source      消息原包
    """

    def args_func(self, message):
        value = message.get('value')
        return (value.get("type"), value.get("from_uin"), value.get("account"),
                message)
    return _register_message_handler(func, args_func, "system_message")


def discu_message_handler(func):
    """ 装饰处理讨论组消息的函数

    处理函数应接收
        did        讨论组id
        from_uin   发送消息的人
        content    消息内容
        source     消息原包
    """

    def args_func(self, message):
        value = message.get("value")
        from_uin = value.get("send_uin")
        did = value.get("did")
        content = self.handle_qq_msg_contents(
            from_uin, value.get("content", []), did, 1)
        return (did, from_uin, content, message)

    return _register_message_handler(func, args_func, "discu_message")


def file_message_handler(func):
    """ 装饰处理文件消息的函数

    处理函数应接受
        from_uin        文件发送人
        to_uin          文件接收人
        lcid            文件sessionid (此处为 session_id 字段)
        guid            文件名称 (此处为 name 字段)
        is_cancel       是否是取消发送
        source          消息源包
    """
    def args_func(self, message):
        value = message.get("value", {})
        return (value.get("from_uin"), value.get("to_uin"),
                value.get("session_id"), value.get("name"),
                value.get("cancel_type", None) == 1,
                message)
    return _register_message_handler(func, args_func, "file_message")


def offline_file_message_handler(func):
    """ 装饰处理离线文件的函数

    处理函数应接收
        from_uin        文件发送人
        to_uin          文件接收人
        lcid            文件sessionid (此处为 session_id 字段)
        count           离线文件数量
        file_infos      文件信息
        source          消息源包
    """
    warnings.warn(u"WebQQ的离线消息并不可靠, 对方可能发送, 但是收不到")

    def args_func(self, message):
        value = message.get("value", {})
        return (value.get("from_uin"), value.get("to_uin"),
                value.get("lcid"), value.get("count"),
                value.get("file_infos"), message)

    return _register_message_handler(func, args_func, "filesrv_transfer")


def check_request(request):
    """ 检查Request参数是否合法, 并返回一个类对象
    """
    if inspect.isclass(request):
        if not issubclass(request, WebQQRequest):
            raise ValueError("Request must be a subclass of WebQQRequest")
    elif isinstance(request, WebQQRequest):
        request = request.__class__
    else:
        raise ValueError(
            "Request must be a subclass or instance of WebQQRequest")

    return request


def register_request_handler(request):
    """ 返回一个装饰器, 用于装饰函数,注册为Request的处理函数
    处理函数需接收两个参数:

        request         本次请求的实例
        response        相应 ~tornado.httpclient.HTTPResponse instance
        data            response.body or dict

    :param request: 请求类或请求实例
    :type request: WebQQRequest or WebQQRequest instance
    :rtype: decorator function
    """
    def wrap(func):
        func._twqq_request = check_request(request)
        return func
    return wrap



#!/usr/bin/env python
# -*- coding:utf-8 -*-
#
#   Author  :   cold
#   E-mail  :   wh_linux@126.com
#   Date    :   13/11/14 13:14:54
#   Desc    :   请求中间件
#
import re
import os
import time
import copy
import json
import random

try:
    import urllib
except ImportError:
    import urllib.parse as urllib

import logging
import tempfile
import threading


from hashlib import md5

try:
    from cStringIO import StringIO
except ImportError:
    from io import StringIO

import pycurl

from tornado.stack_context import ExceptionStackContext
from tornadohttpclient import TornadoHTTPClient

from .requests import check_request, AcceptVerifyRequest
from .requests import WebQQRequest, PollMessageRequest, HeartbeatRequest
from .requests import SessMsgRequest, BuddyMsgRequest, GroupMsgRequest
from .requests import FirstRequest, Login2Request, DiscuMsgRequest
from .requests import FileRequest, LogoutRequset, FriendListRequest
from .requests import GroupMembersRequest

import _hash
import const
import objects

logger = logging.getLogger("twqq")


class RequestHub(object):

    """ 集成Request请求和保存请求值
    :param qid: qq号
    :param pwd: 密码
    :param client: ~twqq.client.Client instance
    """
    SIG_RE = re.compile(r'var g_login_sig=encodeURIComponent\("(.*?)"\);')

    def __init__(self, qid, pwd, client=None, debug=False,
                 handle_msg_image=True):
        self.handle_msg_image = handle_msg_image
        self.http = TornadoHTTPClient()
        self.http.set_user_agent(
            "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36"
            " (KHTML, like Gecko) Ubuntu Chromium/28.0.1500.71 "
            "Chrome/28.0.1500.71 Safari/537.36")
        self.http.validate_cert = False
        self.http.set_global_headers({"Accept-Charset": "UTF-8,*;q=0.5"})
        self.http.debug = debug

        self.qid = qid
        self.__pwd = pwd
        self.client = client

        self.rc = random.randrange(0, 100)
        self.aid = 1003903                                    # aid 固定
        self.clientid = random.randrange(11111111, 99999999)  # 客户端id 随机固定
        self.msg_id = random.randrange(11111111, 99999999)     # 消息id, 随机初始化
        self.daid = 164
        self.login_sig = None
        self.ptwebqq = None
        self.nickname = u"YouWillNeverGetIt"
        self.vfwebqq = None
        self.psessionid = None
        self.stop_poll = False

        # 检查是否验证码的回调
        self.ptui_checkVC = lambda r, v, u: (r, v, u)

        # 是否需要验证码
        self.require_check = None
        self.require_check_time = None

        # 是否开始心跳和拉取消息
        self.poll_and_heart = None
        self.login_time = None
        self.hThread = None

        # 验证图片
        self.checkimg_path = tempfile.mktemp(".jpg")
        self._lock_path = tempfile.mktemp()
        self._wait_path = tempfile.mktemp()

        self.group_sig = {}          # 组签名映射, 用作发送临时消息(sess_message)

        self.message_interval = 0.5  # 消息间隔
        self.last_msg_time = time.time()
        self.last_msg_content = None
        self.last_msg_numbers = 0    # 剩余位发送的消息数量
        WebQQRequest.hub = self
        self.connecting = False

    def connect(self):
        self.connecting = True
        self.load_next_request(FirstRequest())

    def load_next_request(self, request):
        """ 加载下一个请求

        :param request: ~twqq.requests.WebQQRequest instance
        :rtype: ~twqq.requests.WebQQRequest instance
        """
        func = self.http.get if request.method == WebQQRequest.METHOD_GET \
            else self.http.post

        if self.stop_poll and isinstance(request, PollMessageRequest):
            logger.info("检测Poll已停止, 此请求不处理: {0}".format(request))
            return

        kwargs = copy.deepcopy(request.kwargs)
        callback = request.callback if hasattr(request, "callback") and\
            callable(request.callback) else None
        kwargs.update(callback=self.wrap(request, callback))
        kwargs.update(headers=request.headers)
        kwargs.update(delay=request.delay)
        logger.debug("KWARGS: {0}".format(kwargs))

        if request.ready:
            logger.debug("处理请求: {0}".format(request))
            with ExceptionStackContext(request.handle_exc):
                func(request.url, request.params, **kwargs)
        else:
            logger.debug("请求未就绪: {0}".format(request))

        return request

    def handle_pwd(self, r, vcode, huin):
        """ 根据检查返回结果,调用回调生成密码和保存验证码 """
        pwd = md5(md5(self.__pwd).digest() + huin).hexdigest().upper()
        pwd = md5(pwd + vcode).hexdigest().upper()
        return pwd

    def upload_file(self, path):
        """ 上传文件

        :param path: 文件路径
        """
        img_host = "http://dimg.vim-cn.com/"
        curl, buff = self.generate_curl(img_host)
        curl.setopt(pycurl.POST, 1)
        curl.setopt(pycurl.HTTPPOST, [('name', (pycurl.FORM_FILE, path)), ])
        try:
            curl.perform()
            ret = buff.getvalue()
            curl.close()
            buff.close()
        except:
            logger.warn(u"上传图片错误", exc_info=True)
            return u"[图片获取失败]"
        return ret

    def generate_curl(self, url=None, headers=None):
        """ 生成一个curl, 返回 curl 实例和用于获取结果的 buffer
        """
        curl = pycurl.Curl()
        buff = StringIO()

        curl.setopt(pycurl.COOKIEFILE, "cookie")
        curl.setopt(pycurl.COOKIEJAR, "cookie_jar")
        curl.setopt(pycurl.SHARE, self.http._share)
        curl.setopt(pycurl.WRITEFUNCTION, buff.write)
        curl.setopt(pycurl.FOLLOWLOCATION, 1)
        curl.setopt(pycurl.MAXREDIRS, 5)
        curl.setopt(pycurl.TIMEOUT, 3)
        curl.setopt(pycurl.CONNECTTIMEOUT, 3)

        if url:
            curl.setopt(pycurl.URL, url)

        if headers:
            self.set_curl_headers(curl, headers)

        return curl, buff

    def set_curl_headers(self, curl, headers):
        """ 将一个字典设置为 curl 的头
        """
        h = []
        for key, val in headers.items():
            h.append("{0}: {1}".format(key, val))
        curl.setopt(pycurl.HTTPHEADER, h)

    def get_msg_img(self, from_uin, file_path):
        """ 获取聊天信息中的图片
        """
        url = "http://d.web2.qq.com/channel/get_offpic2"
        params = {"clientid": self.clientid, "f_uin": from_uin,
                  "file_path": file_path, "psessionid": self.psessionid}
        url = url + "?" + urllib.urlencode(params)
        headers = {}

        headers = {
            "User-Agent":
            "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36"
            " (KHTML, like Gecko) Ubuntu Chromium/28.0.1500.71 "
            "Chrome/28.0.1500.71 Safari/537.36",
            "Referer":  "http://web2.qq.com/webqq.html"}
        curl, buff = self.generate_curl(url, headers)
        try:
            curl.perform()
        except:
            logger.warn(u"获取聊天图片错误", exc_info=True)
            return u"[图片获取失败]"
        body = buff.getvalue()
        curl.close()
        buff.close()

        path = tempfile.mktemp()
        with open(path, 'w') as f:
            f.write(body)
        return self.upload_file(path)

    def get_group_img(self, gid, from_uin, file_id, server, name, key,
                      _type=0):
        """ 获取群发送的图片
        """
        ip, port = server.split(":")
        url = "http://web2.qq.com/cgi-bin/get_group_pic"
        params = {"type": _type, "fid": file_id, "gid": gid, "pic": name,
                  "rip": ip, "rport": port, "uin": from_uin,
                  "vfwebqq": self.vfwebqq}
        url = url + "?" + urllib.urlencode(params)
        headers = {
            "User-Agent":
            "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36"
            " (KHTML, like Gecko) Ubuntu Chromium/28.0.1500.71 "
            "Chrome/28.0.1500.71 Safari/537.36",
            "Referer":  "http://web2.qq.com/webqq.html"}
        curl, buff = self.generate_curl(url, headers)
        try:
            curl.perform()
        except:
            logger.warn(u"获取群聊天图片错误", exc_info=True)
            return u"[图片获取失败]"
        body = buff.getvalue()
        buff.close()
        curl.close()
        path = tempfile.mktemp()
        with open(path, 'w') as f:
            f.write(body)
        return self.upload_file(path)

    def set_friends(self, data):
        """ 存储好友信息
        """
        self._friends = objects.Friends(data)

    def get_friends(self):
        return self._friends if hasattr(self, "_friends") else None

    def set_groups(self, data):
        self._groups = objects.GroupList(data)

    def get_groups(self):
        return self._groups if hasattr(self, "_groups") else None

    def set_discu(self, data):
        self._discu = objects.DiscuList(data)

    def get_discu(self):
        return self._discu if hasattr(self, "_discu") else None

    def lock(self):
        """ 当输入验证码时锁住
        """
        with open(self._lock_path, 'w'):
            pass

    def get_account(self, uin, _type=1):
        """ 获取好友QQ号
        :param _type: 类型, 1 是好友和讨论组, 4 是群
        """
        # self.load_next_request(QQNumberRequest())
        ret = self.get_friends().get_account(uin)
        if ret:
            return ret

        url = "http://s.web2.qq.com/api/get_friend_uin2"
        params = {"code": "", "t": time.time() * 1000, "tuin": uin,
                  "type": _type, "verifysession": "", "vfwebqq": self.vfwebqq}
        url = url + "?" + urllib.urlencode(params)
        headers = {
            "User-Agent":
            "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36"
            " (KHTML, like Gecko) Ubuntu Chromium/28.0.1500.71 "
            "Chrome/28.0.1500.71 Safari/537.36",
            "Referer":  const.S_REFERER}
        curl, buff = self.generate_curl(url, headers)

        try:
            curl.perform()
            ret = buff.getvalue()
            buff.close()
            data = json.loads(ret)
            curl.close()
        except:
            logger.warn(u"获取QQ号时发生错误", exc_info=True)
            return

        if data.get("retcode") == 0:
            logger.info(u"获取QQ号码成功: {0!r}".format(data))
            ret = data.get("result")
            uin = ret.get("uin")
            account = ret.get("account")
            self.get_friends().set_account(uin, account)
            return account
        logger.warn(u"获取QQ号码失败: {0!r}".format(data))

    def unlock(self):
        """ 解锁
        """
        if os.path.exists(self._lock_path):
            os.remove(self._lock_path)

    def clean(self):
        """ 清除锁住和等待状态
        """
        self.unlock()
        self.unwait()

    def wait(self):
        """ 当没有验证是否需要验证码时等待
        """
        with open(self._wait_path, 'w'):
            pass

    def unwait(self):
        """ 解除等待状态
        """
        if os.path.exists(self._wait_path):
            os.remove(self._wait_path)

    def is_lock(self):
        """ 检测是否被锁住
        """
        return os.path.exists(self._lock_path)

    def is_wait(self):
        """ 检测是否在等待生成验证码
        """
        return os.path.exists(self._wait_path)

    def _hash(self):
        """  获取好友列表时的Hash """
        return _hash.webqq_hash(self.qid, self.ptwebqq)

    def start_poll(self):
        """ 开始心跳和拉取信息
        """
        self.stop_poll = False
        if not self.poll_and_heart:
            self.login_time = time.time()
            logger.info("开始拉取信息")
            self.load_next_request(PollMessageRequest())
            self.poll_and_heart = True
            if self.hThread is None:
                logger.info("开始心跳")
                self.hThread = threading.Thread(target=self._heartbeat)
                self.hThread.setDaemon(True)
                self.hThread.start()

    def _heartbeat(self):
        """ 放入线程的产生心跳
        """
        assert not isinstance(threading.currentThread(), threading._MainThread)
        while 1:
            try:
                self.load_next_request(HeartbeatRequest())
            except:
                pass
            time.sleep(60)

    def make_msg_content(self, content, style):
        """ 构造QQ消息的内容

        :param content: 小心内容
        :type content: str
        :rtype: str
        """
        self.msg_id += 1
        return json.dumps([content,
                           ["font", style]])

    def get_delay(self, content):
        """ 根据消息内容是否和上一条内容相同和未送出的消息数目产生延迟

        :param content: 消息内容
        :rtype: tuple(delay, number)
        """
        MIN = self.message_interval
        delay = 0
        sub = time.time() - self.last_msg_time
        if self.last_msg_numbers < 0:
            self.last_msg_numbers = 0

        # 不足最小间隔就补足最小间隔
        if sub < MIN:
            delay = MIN
            logger.debug(u"间隔 %s 小于 %s, 设置延迟为%s", sub, MIN, delay)

        # 如果间隔是已有消息间隔的2倍, 则清除已有消息数
        # print "sub", sub, "n:", self.last_msg_numbers
        if self.last_msg_numbers > 0 and\
                sub / (MIN * self.last_msg_numbers) > 1:
            self.last_msg_numbers = 0

        # 如果还有消息未发送, 则加上他们的间隔
        if self.last_msg_numbers > 0:
            delay += MIN * self.last_msg_numbers
            logger.info(u"有%s条消息未发送, 延迟为 %s",
                        self.last_msg_numbers, delay)

        n = 1
        # 如果这条消息和上条消息一致, 保险起见再加上一个最小间隔
        if self.last_msg_content == content and sub < MIN:
            delay += MIN
            self.last_msg_numbers += 1
            n = 2

        self.last_msg_numbers += 1
        self.last_msg_content = content

        if delay:
            logger.info(u"有 {1} 个消息未投递将会在 {0} 秒后投递"
                        .format(delay, self.last_msg_numbers))
        # 返回消息累加个数, 在消息发送后减去相应的数目
        return delay, n

    def consume_delay(self, number):
        """ 消费延迟

        :param number: 消费的消息数目
        """
        self.last_msg_numbers -= number
        self.last_msg_time = time.time()

    def get_group_id(self, uin):
        """ 根据组uin获取组的id

        :param uin: 组的uin
        """
        return self.get_groups().get_gid(uin)

    def get_friend_name(self, uin):
        """ 获取好友名称

        :param uin: 好友uin
        """
        return self.get_friends().get_show_name(uin)

    def wrap(self, request, func=None):
        """ 装饰callback

        :param request: ~twqq.requests.WebQQRequest instance
        :param func: 回调函数
        """
        def _wrap(resp, *args, **kwargs):
            data = resp.body
            logger.debug(resp.headers)
            if resp.headers.get("Content-Type") == "application/json":
                data = json.loads(data) if data else {}
            else:
                try:
                    data = json.loads(data)
                except:
                    pass
            if func:
                func(resp, data, *args, **kwargs)

            funcs = self.client.request_handlers.get(
                check_request(request), [])
            for f in funcs:
                f(request, resp, data)

        return _wrap

    def handle_qq_msg_contents(self, from_uin, contents, eid=None, _type=0):
        """ 处理QQ消息内容

        :param from_uin: 消息发送人uin
        :param contents: 内容
        :param eid: 扩展id(群gid, 讨论组did)
        :type contents: list

        """
        content = ""
        for row in contents:
            if isinstance(row, (list)) and len(row) == 2:
                info = row[1]
                if row[0] == "offpic" and self.handle_msg_image:
                    file_path = info.get("file_path")
                    content += self.get_msg_img(from_uin, file_path)

                if row[0] == "cface" and self.handle_msg_image:
                    name = info.get("name")
                    key = info.get("key")
                    file_id = info.get("file_id")
                    server = info.get("server")
                    content += self.get_group_img(eid, from_uin, file_id,
                                                  server, name, key, _type)

            if isinstance(row, (str, unicode)):
                content += row.replace(u"【提示：此用户正在使用Q+"
                                       u" Web：http://web.qq.com/】", "")\
                    .replace(u"【提示：此用户正在使用Q+"
                             u" Web：http://web3.qq.com/】", "")
        return content.replace("\r", "\n").replace("\r\n", "\n")\
            .replace("\n\n", "\n")

    def get_group_member_nick(self, gcode, uin):
        """ 根据组代码和用户uin获取群成员昵称

        :param gcode: 组代码
        :param uin: 群成员uin
        """
        return self.get_groups().get_member_nick(gcode, uin)

    def dispatch(self, qq_source):
        """ 调度QQ消息

        :param qq_source: 源消息包
        """
        if self.stop_poll:
            logger.info("检测Poll已停止, 此消息不处理: {0}".format(qq_source))
            return

        if qq_source.get("retcode") == 0:
            messages = qq_source.get("result")
            logger.info(u"获取消息: {0}".format(messages))
            for m in messages:
                poll_type = m.get("poll_type")
                if poll_type == "buddies_status_change":
                    self.get_friends().set_status(**m.get("value", {}))
                else:
                    funcs = self.client.msg_handlers.get(m.get("poll_type"),
                                                         [])
                    [func(*func._args_func(self, m)) for func in funcs]

    def recv_file(self, guid, lcid, to, callback):
        """ 接收文件

        :param guid: 文件名
        :param lcid: 会话id
        :param to_uin: 发送人uin
        :param callback:  回调, 接收两个参数, 分别是文件名和文件内容
        """
        self.load_next_request(FileRequest(guid, lcid, to, callback))

    def relogin(self):
        """ 被T出或获取登出时尝试重新登录
        """
        self.stop_poll = True
        self.poll_and_heart = None
        self.load_next_request(Login2Request(relogin=True))

    def disconnect(self):
        self.stop_poll = True
        self.poll_and_heart = None
        self.load_next_request(LogoutRequset())

    def send_sess_msg(self, qid, to_uin, content, style=const.DEFAULT_STYLE):
        """ 发送临时消息

        :param qid: 发送临时消息的qid
        :param to_uin: 消息接收人
        :param content: 消息内容
        :rtype: Request instance
        """
        return self.load_next_request(SessMsgRequest(qid, to_uin, content,
                                                     style))

    def send_group_msg(self, group_uin, content, style=const.DEFAULT_STYLE):
        """ 发送群消息

        :param group_uin: 组的uin
        :param content: 消息内容
        :rtype: Request instance
        """
        return self.load_next_request(GroupMsgRequest(group_uin, content,
                                                      style))

    def send_discu_msg(self, did, content, style=const.DEFAULT_STYLE):
        """ 发送讨论组消息

        :param did: 讨论组id
        :param content: 内容
        """
        return self.load_next_request(DiscuMsgRequest(did, content, style))

    def send_buddy_msg(self, to_uin, content, style=const.DEFAULT_STYLE):
        """ 发送好友消息

        :param to_uin: 消息接收人
        :param content: 消息内容
        :rtype: Request instance
        """
        return self.load_next_request(BuddyMsgRequest(to_uin, content, style))

    def send_msg_with_markname(self, markname, content):
        """ 使用备注名发送消息

        :param markname: 备注名
        :param content: 消息内容
        :rtype: None or Request instance
        """
        uin = self.get_friends().get_uin_from_mark(markname)
        if not uin:
            return
        return self.send_buddy_msg(uin, content)

    def accept_verify(self, uin, account, markname=""):
        """ 同意验证请求

        :param  uin: 请求人uin
        :param account: 请求人账号
        :param markname: 添加后的备注
        """
        return self.load_next_request(AcceptVerifyRequest(uin, account,
                                                          markname))

    def refresh_friend_info(self):
        self.load_next_request(FriendListRequest(manual=True))

    def refresh_group_info(self, _id):
        """ 手动刷新某个群的信息

        :param _id: 对应群生成的唯一id
        """
        gcode, _type = objects.UniqueIds.get(int(_id))
        if gcode is None or _type is None:
            return False, u"没有找到对象"

        if _type != objects.UniqueIds.T_GRP:
            return False, u"该对象不是群"

        self.load_next_request(GroupMembersRequest(gcode))
        return True, self.get_groups().get_group_name(gcode)



#!/usr/bin/env python
# -*- coding:utf-8 -*-
#
#   Author  :   cold
#   E-mail  :   wh_linux@126.com
#   Date    :   13/09/05 11:16:58
#   Desc    :
#
import twqq
from setuptools import setup

requires = ["tornado", "pycurl", "tornadohttpclient"]

packages = ["twqq"]

entry_points = {
}


setup(
    name = "twqq",
    version = twqq.__version__,
    description = 'An asynchronous webqq client library based on tornado',
    long_description = open("README.rst").read(),
    author = 'cold',
    author_email = 'wh_linux@126.com',
    url = 'http://www.linuxzen.com',
    license = 'Apache 2.0',
    platforms = 'any',
    packages = packages,
    package_data = {
    },
    entry_points = entry_points,
    install_requires = requires,
    classifiers=['Development Status :: 3 - Alpha',
                 'Environment :: Console',
                 "Intended Audience :: Developers",
                 'License :: OSI Approved :: Apache Software License',
                 'Topic :: Internet :: WWW/HTTP',
                 'Programming Language :: Python :: 2.7',
                 ],
)
